import time
import os
import base64

def generate_ulid():
    # Get the current timestamp in milliseconds
    timestamp = int(time.time() * 1000)
    # Generate 10 bytes of random data
    random_data = os.urandom(10)
    # Encode the timestamp and random data using Crockford's Base32 encoding
    ulid = base64.b32encode(timestamp.to_bytes(6, 'big') + random_data).decode('utf-8').replace('=', '')
    return ulid



# -- MongoDB Below 

from pymongo import MongoClient
from bson.objectid import ObjectId
import json
from contextlib import contextmanager
from .utils import generate_ulid

class Database:
    def __init__(self, connection_string):
        self.client = MongoClient(connection_string)
        self.db = self.client.get_database()

    @contextmanager
    def get_collection(self, collection_name):
        try:
            yield self.db[collection_name]
        finally:
            pass

class CloudBase:
    def __init__(self, connection_string):
        self.db = Database(connection_string)

    def __call__(self, collection_name):
        return Collection(self.db, collection_name)

class Collection:
    def __init__(self, db, collection_name):
        self.db = db
        self.collection_name = collection_name

    def fetchall(self, query=None):
        with self.db.get_collection(self.collection_name) as collection:
            if query is None:
                results = collection.find()
            else:
                results = collection.find(self._transform_query(query))
            return [self._transform_document(doc) for doc in results]

    def fetch(self, query=None):
        return self.fetchall(query)

    def get(self, id):
        with self.db.get_collection(self.collection_name) as collection:
            result = collection.find_one({"_id": id})
            return self._transform_document(result) if result else None

    def put(self, data):
        id = str(generate_ulid())
        with self.db.get_collection(self.collection_name) as collection:
            data["_id"] = id
            collection.insert_one(data)
        return {"id": id, "msg": "success"}

    def update(self, query, id):
        with self.db.get_collection(self.collection_name) as collection:
            result = collection.update_one({"_id": id}, {"$set": query})
            if result.matched_count == 0:
                collection.insert_one({"_id": id, **query})

    def delete(self, id):
        with self.db.get_collection(self.collection_name) as collection:
            collection.delete_one({"_id": id})

    def truncate(self):
        with self.db.get_collection(self.collection_name) as collection:
            collection.delete_many({})

    def drop(self):
        with self.db.get_collection(self.collection_name) as collection:
            collection.drop()

    def _transform_query(self, query):
        transformed = {}
        for key, value in query.items():
            if "?contains" in key:
                field = key.split("?")[0]
                transformed[field] = {"$regex": value, "$options": "i"}
            else:
                transformed[key] = value
        return transformed

    def _transform_document(self, doc):
        if doc is None:
            return None
        doc['id'] = doc.pop('_id')
        return doc


# -- Postgres Below 

# from sqlalchemy import create_engine, Column, String, JSON, Table as SQLATable, MetaData
# from sqlalchemy.orm import sessionmaker, declarative_base
# from sqlalchemy.exc import SQLAlchemyError
# from sqlalchemy import inspect
# import json
# from contextlib import contextmanager
# from .utils import generate_ulid

# Base = declarative_base()
# # this is latest

# class Database:
#     def __init__(self, db_url):
#         self.engine = create_engine(db_url)
#         self.SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
#         self.metadata = MetaData()

#     @contextmanager
#     def get_session(self):
#         session = self.SessionLocal()
#         try:
#             yield session
#         finally:
#             session.close()

# class CloudBase:
#     def __init__(self, db_url):
#         self.db = Database(db_url)

#     def __call__(self, table_name):
#         return Table(self.db, table_name)

# class Table:
#     def __init__(self, db, table_name):
#         self.db = db
#         self.table_name = table_name
#         self.table = self.get_or_create_table()

#     def get_or_create_table(self):
#         inspector = inspect(self.db.engine)
#         if not inspector.has_table(self.table_name):
#             table = SQLATable(self.table_name, self.db.metadata,
#                 Column('id', String, primary_key=True),
#                 Column('data', JSON)
#             )
#             self.db.metadata.create_all(self.db.engine)
#         else:
#             table = SQLATable(self.table_name, self.db.metadata, autoload_with=self.db.engine)
#         return table

#     def parse_result(self, row):
#         try:
#             data = json.loads(row.data) if isinstance(row.data, str) else row.data
#             return {'id': row.id, **data}
#         except json.JSONDecodeError:
#             print(f"Warning: Could not parse JSON for id {row.id}. Returning raw data.")
#             return {'id': row.id, 'data': row.data}

#     def fetchall(self, query=None):
#         with self.db.get_session() as session:
#             results = session.query(self.table).all()
#             parsed_results = [self.parse_result(row) for row in results]
#             if query is None:
#                 return parsed_results
#             else:
#                 return self.filter_results(parsed_results, query)

#     def fetch(self, query=None):
#         return self.fetchall(query)

#     def filter_results(self, results, query):
#         filtered_results = []
#         for result in results:
#             match = True
#             for key, value in query.items():
#                 if "?contains" in key:
#                     field = key.split("?")[0]
#                     if value.lower() not in str(result.get(field, "")).lower():
#                         match = False
#                         break
#                 else:
#                     if result.get(key) != value:
#                         match = False
#                         break
#             if match:
#                 filtered_results.append(result)
#         return filtered_results

#     def put(self, data):
#         id = str(generate_ulid())
#         with self.db.get_session() as session:
#             new_row = self.table.insert().values(id=id, data=data)
#             session.execute(new_row)
#             session.commit()
#         return {"id": id, "msg": "success"}

#     def update(self, query, id):
#         with self.db.get_session() as session:
#             update_stmt = self.table.update().where(self.table.c.id == id).values(data=query)
#             session.execute(update_stmt)
#             session.commit()

#     def delete(self, id):
#         with self.db.get_session() as session:
#             delete_stmt = self.table.delete().where(self.table.c.id == id)
#             session.execute(delete_stmt)
#             session.commit()

#     def truncate(self):
#         with self.db.get_session() as session:
#             session.execute(self.table.delete())
#             session.commit()

#     def drop(self):
#         self.table.drop(self.db.engine)

#     def get(self, id):
#         with self.db.get_session() as session:
#             result = session.query(self.table).filter(self.table.c.id == id).first()
#             return self.parse_result(result) if result else None
        
# # --------------------------------------------------------------------------------------------------------------------------------
# # # # # # Initialize the database connection (example with SQLite)

# # db_url = "postgresql://neondb_owner:5JVIgPtBlk6R@ep-fancy-mountain-a7mulxw3.ap-southeast-2.aws.neon.tech/neondb?sslmode=require"
# # db = CloudBase(db_url)

# # # Access the 'users' table
# # users = db('users')

# # # Example: Insert a user
# # new_user = users.put({"name": "John Doe", "email": "john@example.com"})
# # print("Inserted user:", new_user)

# # # Get the user by id
# # user_id = new_user['id']
# # retrieved_user = users.get(user_id)
# # print("Retrieved user:", retrieved_user)

# # # Fetch all users
# # all_users = users.fetch()
# # print("All users:", all_users)

# # # Update a user
# # users.update({"name": "John Updated", "email": "john.updated@example.com"}, user_id)
# # updated_user = users.get(user_id)
# # print("Updated user:", updated_user)

# # # Delete a user
# # users.delete(user_id)
# # deleted_user = users.get(user_id)
# # print("Deleted user (should be None):", deleted_user)


import sqlite3
import json
from contextlib import contextmanager
from .utils import generate_ulid

class Database:
    def __init__(self, db_name):
        self.db_name = db_name
        self.conn = None
        self.cursor = None

    @contextmanager
    def get_conn(self):
        db_path = self.db_name
        self.conn = sqlite3.connect(db_path)
        try:
            yield self.conn
        finally:
            self.conn.close()

    @contextmanager
    def get_cursor(self):
        with self.get_conn() as conn:
            self.cursor = conn.cursor()
            try:
                yield self.cursor
            finally:
                self.cursor.close()

class LocalBase:
    def __init__(self, db_name):
        self.db = Database(db_name)

    def __call__(self, table_name):
        return Table(self.db, table_name)

class Table:
    def __init__(self, db, table_name):
        self.db = db
        self.table_name = table_name

    def get(self, id):
        with self.db.get_cursor() as cursor:
            if not self.table_exists(cursor):
                return None
            cursor.execute(f"SELECT id, data FROM {self.table_name} WHERE id = %s", (id,))
            result = cursor.fetchone()
            if result is None:
                return None
            return self.parse_result(result[0], result[1])

    def fetchall(self, query=None):
        with self.db.get_cursor() as cursor:
            if not self.table_exists(cursor):
                return []
            cursor.execute(f"SELECT id, data FROM {self.table_name}")
            results = cursor.fetchall()
            parsed_results = [{'id': id, **json.loads(data)} for id, data in results]
            if query is None:
                return parsed_results
            else:
                filtered_results = []
                for result in parsed_results:
                    for key, value in query.items():
                        if "?contains" in key:
                            field = key.split("?")[0]
                            if value.lower() in result.get(field, "").lower():
                                filtered_results.append(result)
                                break
                        else:
                            if result.get(key) == value:
                                filtered_results.append(result)
                                break
                return filtered_results

    def fetch(self, query=None):
        with self.db.get_cursor() as cursor:
            if not self.table_exists(cursor):
                return []
            cursor.execute(f"SELECT id, data FROM {self.table_name}")
            results = cursor.fetchall()
            parsed_results = [{'id': id, **json.loads(data)} for id, data in results]
            if query is None:
                return parsed_results
            else:
                filtered_results = []
                for result in parsed_results:
                    match = True
                    for key, value in query.items():
                        if "?contains" in key:
                            field = key.split("?")[0]
                            if value.lower() not in result.get(field, "").lower():
                                match = False
                                break
                        else:
                            if result.get(key) != value:
                                match = False
                                break
                    if match:
                        filtered_results.append(result)
                return filtered_results

    def put(self, data):
        id = str(generate_ulid())
        data_json = json.dumps(data)
        with self.db.get_cursor() as cursor:
            cursor.execute(f"CREATE TABLE IF NOT EXISTS {self.table_name} (id TEXT PRIMARY KEY, data TEXT)")
            cursor.execute(f"INSERT INTO {self.table_name} VALUES (?, ?)", (id, data_json))
            cursor.connection.commit()
        return { "id" : id, "msg" : "success" }    

    def update(self, query, id):
        data_json = json.dumps(query)
        with self.db.get_cursor() as cursor:
            cursor.execute(f"CREATE TABLE IF NOT EXISTS {self.table_name} (id TEXT PRIMARY KEY, data TEXT)")
            cursor.execute(f"SELECT id FROM {self.table_name} WHERE id = ?", (id,))
            if cursor.fetchone() is None:
                cursor.execute(f"INSERT INTO {self.table_name} VALUES (?, ?)", (id, data_json))
            else:
                cursor.execute(f"UPDATE {self.table_name} SET data = ? WHERE id = ?", (data_json, id))
            cursor.connection.commit()

    def delete(self, id):
        with self.db.get_cursor() as cursor:
            if not self.table_exists(cursor):
                return []
            cursor.execute(f"DELETE FROM {self.table_name} WHERE id = ?", (id,))
            cursor.connection.commit()

    def truncate(self):
        with self.db.get_cursor() as cursor:
            if not self.table_exists(cursor):
                return []
            cursor.execute(f"DELETE FROM {self.table_name}")
            cursor.connection.commit()

    def drop(self):
        with self.db.get_cursor() as cursor:
            if not self.table_exists(cursor):
                return []
            cursor.execute(f"DROP TABLE {self.table_name}")
            cursor.connection.commit()

    def table_exists(self, cursor):
        cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name=?", (self.table_name,))
        return cursor.fetchone() is not None



# # from odeta import cloudbase
# from cloudbase import CloudBase as cloudbase

# # Initialize the database with the table name directly
# db = cloudbase('my_table')

# # Insert data
# data = {
#     'name': 'John Doe',
#     'age': 30,
#     'email': 'john.doe@example.com'
# }
# response = db.put(data)
# print(response)  # {'id': 'generated_ulid', 'msg': 'success'}

# # Fetch all records
# all_records = db.fetchall()
# print(all_records)

# # Fetch records with a specific query
# filtered_records = db.fetch({'name?contains': 'John'})
# print(filtered_records)

# # # Update a record by its ID
# # update_data = {
# #     'name': 'Jane Doe',
# #     'age': 25,
# #     'email': 'jane.doe@example.com'
# # }
# # db.update(update_data, 'record_id')

# # # Delete a record by its ID
# # db.delete('record_id')

# # # Truncate the table
# # db.truncate()

# # # Drop the table
# # db.drop()




# import boto3
# from botocore.exceptions import NoCredentialsError, PartialCredentialsError, NoRegionError, WaiterError

# # Replace these with your own credentials
# AWS_ACCESS_KEY_ID = 'AKIARQUMPKD3FKM7KH7Y'
# AWS_SECRET_ACCESS_KEY = 'RC7iw5SIQFsMut43lOTApwPTO3bGSTY2Lj4oDE1X'
# AWS_REGION = 'ap-south-1'  # e.g., 'us-west-2'

# # Create a session using your credentials
# try:
#     session = boto3.Session(
#         aws_access_key_id=AWS_ACCESS_KEY_ID,
#         aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
#         region_name=AWS_REGION
#     )
#     print(f"Session created with region: {session.region_name}")
# except NoRegionError:
#     print("No region specified or region not recognized.")
# except Exception as e:
#     print(f"Error creating session: {e}")

# # Create a DynamoDB client
# try:
#     dynamodb = session.resource('dynamodb', region_name=AWS_REGION)
#     print(f"DynamoDB resource created with region: {AWS_REGION}")
# except NoRegionError:
#     print("No region specified or region not recognized.")
# except Exception as e:
#     print(f"Error creating DynamoDB resource: {e}")

# # Create a table
# def create_table():
#     try:
#         table = dynamodb.create_table(
#             TableName='TestTable',
#             KeySchema=[
#                 {
#                     'AttributeName': 'id',
#                     'KeyType': 'HASH'  # Partition key
#                 }
#             ],
#             AttributeDefinitions=[
#                 {
#                     'AttributeName': 'id',
#                     'AttributeType': 'N'
#                 }
#             ],
#             ProvisionedThroughput={
#                 'ReadCapacityUnits': 5,
#                 'WriteCapacityUnits': 5
#             }
#         )
#         print("Table status:", table.table_status)

#         # Wait for the table to be created
#         table.meta.client.get_waiter('table_exists').wait(TableName='TestTable')
#         print("Table created successfully.")

#     except Exception as e:
#         print(f"Error creating table: {e}")

# # Insert an item into the table
# def insert_item():
#     try:
#         table = dynamodb.Table('TestTable')
#         response = table.put_item(
#             Item={
#                 'id': 1,
#                 'name': 'Test Item',
#                 'description': 'This is a test item'
#             }
#         )
#         print("PutItem succeeded:", response)
#     except Exception as e:
#         print(f"Error inserting item: {e}")

# # Get an item from the table
# def get_item():
#     try:
#         table = dynamodb.Table('TestTable')
#         response = table.get_item(
#             Key={
#                 'id': 1
#             }
#         )
#         item = response.get('Item')
#         print("GetItem succeeded:", item)
#     except Exception as e:
#         print(f"Error getting item: {e}")

# # Delete the table
# def delete_table():
#     try:
#         table = dynamodb.Table('TestTable')
#         table.delete()
#         print("Table deletion initiated")
#     except Exception as e:
#         print(f"Error deleting table: {e}")

# if __name__ == "__main__":
#     try:
#         create_table()
#         insert_item()
#         get_item()
#         # delete_table()
#     except NoCredentialsError:
#         print("Credentials not available")
#     except PartialCredentialsError:
#         print("Incomplete credentials provided")
#     except NoRegionError:
#         print("No region specified or region not recognized.")
#     except Exception as e:
#         print(f"An error occurred: {e}")
